""" This module define the Trade class used to manage positions.
"""
import logging
from typing import Any, Dict, List, Optional, Union, TYPE_CHECKING

from estrade.mixins.ref_mixin import RefMixin
from estrade.exceptions import TradeException
from estrade.observer import Observable
from estrade.stop_limit import StopLimitMixinAbsolute, StopLimitMixinRelative
from estrade.tick import Tick

if TYPE_CHECKING:
    from estrade.epic import Epic
    from estrade.strategy import Strategy
    from estrade.trade_manager import TradeManager


logger = logging.getLogger(__name__)


class TradeClose:
    """
    Manage a trade close

    Arguments:
        tick: tick to close trade with
        quantity: quantity to close on trade
        result: result of trade close
        reason: reason for closing
    """

    def __init__(self, tick: Tick, quantity: int, result: float, reason: str) -> None:
        self.tick = tick
        self.quantity = quantity
        self.result = result
        self.reason = reason

    def to_json(self) -> Dict[str, Any]:
        """
        Convert trade close to dict (mainly used for reporting)

        Returns:
            dictionary formatted close
        """
        return {
            'tick': self.tick.to_json(True),
            'quantity': self.quantity,
            'reason': self.reason,
            'result': self.result,
        }


class Trade(RefMixin, Observable):
    """
    The Trade class manage trades (=market positions)

    Arguments:
        quantity: open trade quantities
        direction: trade direction (`BUY`/`1` or `SELL`/`-1`)
        epic: trade epic to open trade on
        trade_manager: trade manager holding this trade
        strategy: strategy that generated this trade
        ref: trade name reference
        stop_absolute: trade stop in absolute value
        stop_relative: trade stop in relative value
        limit_absolute: trade target profit in absolute value
        limit_relative: trade target profit in relative value
        meta: free of use dictionary

    """

    def __init__(
        self,
        quantity: int,
        direction: Union[str, int],
        trade_manager: 'TradeManager',
        strategy: 'Strategy',
        epic: 'Epic',
        ref: Optional[str] = None,
        stop_absolute: Optional[Union[int, float]] = None,
        stop_relative: Optional[Union[int, float]] = None,
        limit_absolute: Optional[Union[int, float]] = None,
        limit_relative: Optional[Union[int, float]] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> None:
        # set ref with parent class
        RefMixin.__init__(self, ref)

        # set class as observable so it can fire events
        # TODO : fire events on trade update
        Observable.__init__(self)

        # set trade manager
        self.trade_manager: 'TradeManager' = trade_manager

        # set strategy
        self.strategy: 'Strategy' = strategy

        self.direction = direction
        self.inital_quantities = abs(quantity)
        self.opened_quantity = abs(quantity)
        self.closed_quantity = 0
        self.closed = False

        # init empty list of ticks received by trade
        open_tick = epic.ticks[-1]

        if open_tick is None:
            raise TradeException('Impossible to define trade open tick.')
        elif not epic.tradeable:
            raise TradeException(
                'Open tick epic %s is not tradeable ' % open_tick.epic.ref
            )
        self.open_tick = open_tick
        self.last_tick = open_tick
        self.high_tick = open_tick
        self.low_tick = open_tick
        self.opened_result = open_tick.spread * -1
        self.closed_result = 0
        self.result = self.opened_result

        # init empty list of closes
        self.closes: List[TradeClose] = []

        # set max gain and max losses
        self.max_gain = float('inf') * -1
        self.max_loss = float('inf')

        # init stop/limit from params
        self._stop = None
        if stop_relative:
            self.set_stop(stop_value=stop_relative, relative=True)
        elif stop_absolute:
            self.set_stop(stop_value=stop_absolute, relative=False)

        self._limit = None
        if limit_relative:
            self.set_limit(limit_value=limit_relative, relative=True)
        elif limit_absolute:
            self.set_limit(limit_value=limit_absolute, relative=False)

        self.on_new_tick(open_tick)

        self.meta = {} if not meta or not isinstance(meta, dict) else meta

        logger.info('%s' % repr(self))

    ##################################################
    # DIRECTION GET/SET
    ##################################################
    @property
    def direction(self) -> int:
        """
        Returns:
            trade direction, `1` for a BUY and `-1` for a sell.
        """
        return self._direction

    @direction.setter
    def direction(self, direction: Union[str, int]) -> None:
        """
        Set trade direction

        Arguments:
            direction: value in `[1, 'BUY']` to open a buy or value in `[-1, 'SELL']`
                to open a sell
        """
        if direction:
            if direction == 1 or (
                isinstance(direction, str) and direction.upper() == 'BUY'
            ):
                self._direction = 1
                return
            elif direction == -1 or (
                isinstance(direction, str) and direction.upper() == 'SELL'
            ):
                self._direction = -1
                return
        raise TradeException(f'Invalid trade direction: {direction}')

    ##################################################
    # STOP/LIMIT HELPERS
    ##################################################
    def _set_stop_limit(
        self, type_: str, value: Union[int, float], relative: bool
    ) -> Union[StopLimitMixinAbsolute, StopLimitMixinRelative]:
        """
        Set a stop or a limit on trade

        Arguments:
            type_: `STOP` or `LIMIT`
            value: stop/limit value
            relative: is the stop/limit value is relative or absolute

        Returns:
            Stop/Limit object
        """
        if relative:
            # set stop/limit as relative
            return StopLimitMixinRelative(type_=type_, trade=self, value=value,)

        # set stop/limit as an absolute stop
        return StopLimitMixinAbsolute(type_=type_.lower(), trade=self, value=value)

    ##################################################
    # STOP GET/SET
    ##################################################
    @property
    def stop(self) -> Optional[Union[StopLimitMixinRelative, StopLimitMixinAbsolute]]:
        """
        Returns:
            Stop/Limit object
        """
        return self._stop

    @stop.setter
    def stop(self, stop: Any) -> None:
        """
        Prevent set of stop directly as stop set requires to define if the stop is
        relative (use set_stop method)

        Argument:
            Stop value

        !!! warning
            Do not use this method, use the set_stop

        """
        raise NotImplementedError(
            'Impossible to directly set stop on trade, use Trade.set_stop method'
        )

    def set_stop(self, stop_value: Union[int, float], relative: bool = False) -> None:
        """
        Stop setter

        Arguments:
            stop_value: stop value
            relative: is the stop value relative to trade open tick or absolute
        """
        self._stop = self._set_stop_limit(
            type_='stop', value=stop_value, relative=relative
        )

    ##################################################
    # LIMIT GET/SET
    ##################################################
    @property
    def limit(self) -> Optional[Union[StopLimitMixinAbsolute, StopLimitMixinRelative]]:
        """
        Returns:
            Trade Stop/Limit object
        """
        return self._limit

    @limit.setter
    def limit(self, limit: Any) -> None:
        """
        Prevent set of limit directly as limit set requires to define if the limit is
        relative (use set_limit method)

        Arguments:
            limit: limit value

        !!! warning
            Do not use this method, use `Trade.set_limit()`

        """
        raise NotImplementedError(
            'Impossible to directly set limit on trade, use Trade.set_limit method'
        )

    def set_limit(self, limit_value: Union[int, float], relative: bool = False) -> None:
        """
        Limit setter

        Arguments:
            limit_value: value of the target profit
            relative: is the value provided relative to trade open tick or absolute
        """
        self._limit = self._set_stop_limit(
            type_='limit', value=limit_value, relative=relative
        )

    ##################################################
    # EPIC GETTERS
    ##################################################
    @property
    def epic(self) -> 'Epic':
        """
        get trade epic

        Returns:
            Trade Epic
        """
        return self.open_tick.epic

    ##################################################
    # OPEN GETTERS
    ##################################################
    @property
    def open(self) -> float:
        """
        Returns:
            trade open value (value of open tick)
        """
        if self.direction > 0:
            return self.open_tick.ask

        return self.open_tick.bid

    @property
    def high(self) -> float:
        """
        Returns:
            trade higher tick value
        """
        return self.high_tick.bid if self.direction > 0 else self.high_tick.ask

    @property
    def low(self) -> float:
        """
        Returns:
            trade lower tick value
        """
        return self.high_tick.bid if self.direction > 0 else self.high_tick.ask

    ##################################################
    # QUANTITIES GETTERS
    ##################################################
    @property
    def quantity(self) -> int:
        """
        Returns:
            current opened quantities
        """
        return self.opened_quantity

    ##################################################
    # CLOSE
    ##################################################
    def close(
        self, tick: Tick = None, quantity: int = None, close_reason: str = 'manual'
    ) -> None:
        """
        Total or partial close of trade.

        Arguments:
            tick: tick to close trade with (last epic tick if not provided)
            quantity: quantity to close (all closed if not provided)
            close_reason: justification of the trade close (for reporting use)

        """
        # define quantities to close
        if quantity and quantity > self.opened_quantity:
            raise TradeException(
                f'Impossible to close {quantity} '
                f'(more than remaining quantity {self.opened_quantity}'
            )
        if not quantity:
            quantity = self.opened_quantity

        # define tick to use for close
        if not tick:
            tick = self.epic.ticks[-1]

        # prevent close when epic is not tradeable
        if not tick.epic.tradeable:
            logger.error(
                'Impossible to close trade because epic %s is not tradeable.'
                % tick.epic.ref
            )
            return

        # calculate result of this close
        close_result = self._calculate_result(close_tick=tick, quantity=quantity)
        self.closed_result += close_result

        # add a new close object to list of closed.
        self.closes.append(
            TradeClose(
                tick=tick, quantity=quantity, result=close_result, reason=close_reason,
            )
        )
        self.closed_quantity += quantity
        self.opened_quantity -= quantity
        if self.opened_quantity <= 0:
            self.result = self.closed_result
            self.closed = True
        else:
            self.result = self._calculate_result(tick) + self.closed_result

        logger.info(
            '%s : Close %d @ %f > %f (max loss: %f, max_gain: %f)'
            % (
                tick.datetime,
                quantity,
                tick.value,
                close_result,
                self.max_loss,
                self.max_gain,
            )
        )

    ##################################################
    # RESULT
    ##################################################
    def _calculate_result(self, close_tick: Tick = None, quantity: int = None) -> float:
        """
        Calculate trade result for close.

        Returns:
            The trade result
        """
        if not close_tick:
            close_tick = self.last_tick
        if not quantity:
            quantity = self.opened_quantity

        if self.direction > 0:
            return round((close_tick.bid - self.open_tick.ask) * abs(quantity), 2)
        else:
            return round((self.open_tick.bid - close_tick.ask) * abs(quantity), 2)

    ##################################################
    # HANDLE NEW TICK
    ##################################################
    def on_new_tick(self, tick: Tick):
        """
        Method triggered when a new tick is received by trade_manager.

        Arguments:
            tick: new tick received

        """
        if not self.closed:
            self.last_tick = tick
            if tick.value > self.high_tick.value:
                self.high_tick = tick
            if tick.value < self.low_tick.value:
                self.low_tick = tick

            # update max gain and max loss
            self.opened_result = self._calculate_result(tick)
            self.result = self.closed_result + self.opened_result

            self.update_max_loss_gains(self.result)

            # check stop/limit
            logger.debug(
                'Check if tick value reached stop (%s) or limit (%s)'
                % (self.stop, self.limit)
            )
            if self.stop and self.stop.check(tick):
                if self.trade_manager:
                    self.trade_manager.close_trade(trade=self, close_reason='stop')
                else:
                    self.close(close_reason='stop')
            elif self.limit and self.limit.check(tick):
                if self.trade_manager:
                    self.trade_manager.close_trade(trade=self, close_reason='limit')
                else:
                    self.close(close_reason='limit')

    def update_max_loss_gains(self, current_result: float) -> None:
        """
        Update trade max gain and max losses.

        Arguments:
            current_result: current trade result
        """
        if current_result < self.max_loss:
            self.max_loss = self.result
        elif current_result > self.max_gain:
            self.max_gain = self.result

    ##################################################
    # REPORTING
    ##################################################
    @property
    def to_json(self):
        return {
            'ref': self.ref,
            'open_at': self.open_tick.datetime,
            'open': self.open_tick.value,
            'open_quantity': abs(self.inital_quantities),
            'direction': self.direction,
            'stop_rel': self.stop.relative_value if self.stop else None,
            'stop_abs': self.stop.absolute_value if self.stop else None,
            'limit_rel': self.limit.relative_value if self.limit else None,
            'limit_abs': self.limit.absolute_value if self.limit else None,
            'closed_quantity': abs(self.closed_quantity),
            'closed': self.closed,
            'closes': [c.to_json() for c in self.closes],
            'result': self.result,
            'max_loss': self.max_loss,
            'max_gain': self.max_gain,
            **self.meta,
        }

    @property
    def json_headers(self):
        base_headers = [
            'ref',
            'open_at',
            'open',
            'open_quantity',
            'direction',
            'stop_rel',
            'stop_abs',
            'limit_rel',
            'limit_abs',
            'closed_quantity',
            'closed',
            'closes',
            'result',
            'max_loss',
            'max_gain',
        ]
        base_headers.extend(list(self.meta.keys()))
        return base_headers

    def __repr__(self):
        repr_string = (
            f'{self.open_tick.datetime}: '
            f'Trade for {self.inital_quantities} @ {self.open} '
            f'with ref: {self.ref} '
            f'on strategy {self.strategy.ref if self.strategy else None}'
        )
        if self.stop:
            repr_string = (
                f'{repr_string}, '
                f'stop absolute: {self.stop.absolute_value}, '
                f'stop relative {self.stop.relative_value}'
            )
        if self.limit:
            repr_string = (
                f'{repr_string}, '
                f'limit absolute: {self.limit.absolute_value}, '
                f'limit relative {self.limit.relative_value}'
            )
        return repr_string



import logging
from datetime import datetime
from typing import Dict, List, Optional, Union, TYPE_CHECKING

import arrow

from estrade.exceptions import TradeManagerException
from estrade.observer import Observable
from estrade.trade import Trade

if TYPE_CHECKING:
    from estrade.epic import Epic
    from estrade.market import Market
    from estrade.tick import Tick
    from estrade.strategy import Strategy

logger = logging.getLogger(__name__)


class TradeManager(Observable):
    """
    A trade manager handle trades created by strategies. A trade Manager is
    attached to a Market object.

    Arguments:
        market: Parent market object

    """

    def __init__(self, market=None):
        # init list of trades
        self.trades: List[Trade] = []
        self.open_trades: List[Trade] = []
        self.closed_trades: List[Trade] = []
        self.strategy_trades: Dict[str, Dict[str : List[Trade]]] = {}

        # init market
        self.market: Market = market

        # set trade manager as observable so it can fire events
        Observable.__init__(self)
        logger.debug(
            'New trade manager created with %d open trades and %d closed trades'
            % (len(self.open_trades), len(self.closed_trades))
        )
        logger.debug('strategy trades: %s' % self.strategy_trades)

    ##################################################
    # Market
    ##################################################
    def add_strategies(self):
        logger.debug('add strategies to trade manager')
        for strategy in self.market.strategies:
            if strategy.ref not in self.strategy_trades:
                logger.debug('add strategy %s to trade manager' % strategy.ref)
                self.strategy_trades[strategy.ref] = {'opened': [], 'closed': []}

    ##################################################
    # Trades
    ##################################################
    def get_trade_by_ref(self, ref: str) -> Optional[Trade]:
        """
        Find a trade by its ref

        Arguments:
            ref: reference of the trade to fetch

        Returns:
            Trade with this reference
        """
        for trade in reversed(self.trades):
            if trade.ref == ref:
                return trade
        return None

    ##################################################
    # OPEN TRADE
    ##################################################
    def open_trade(self, *args, **kwargs) -> Optional[Trade]:
        """
        This method create a new Trade object.

        Arguments:
            args: see `estrade.Trade.open_trade` arguments
            kwargs: see `estrade.Trade.open_trade` arguments

        """
        logger.debug('Open new trade')
        self.fire('trade_manager_before_new_trade', *args, **kwargs)

        new_trade = Trade(*args, trade_manager=self, **kwargs)
        new_trade = self.market.provider.open_trade(new_trade)
        if not new_trade:
            logger.error(
                'New trade creation failed, do not add it to trade manager trades.'
            )
            return None

        self.trades.append(new_trade)
        self.open_trades.append(new_trade)
        if new_trade.strategy:
            self.strategy_trades[new_trade.strategy.ref]['opened'].append(new_trade)

        self.fire('trade_manager_after_open_trade', trade=new_trade)

        return new_trade

    @property
    def open_buys(self) -> List[Trade]:
        """
        Returns:
            List of all opened buy trades
        """
        return [trade for trade in self.open_trades if trade.direction > 0]

    @property
    def open_sells(self) -> List[Trade]:
        """
        Returns:
            List all opened sell trades
        """
        return [trade for trade in self.open_trades if trade.direction < 0]

    ##################################################
    # ON NEW TICK EVENT
    ##################################################
    def on_new_tick(self, tick: 'Tick') -> None:
        """
        On new tick, dispatch tick to all opened trades (to update their result)

        Arguments:
            tick: new tick received.
        """
        for trade in reversed(self.open_trades):
            if tick.epic == trade.epic:
                trade.on_new_tick(tick)

    def on_new_ticks_high_low(self, epic: 'Epic', high: 'Tick', low: 'Tick') -> None:
        """
        This method is used in candle mode to apply :

         - candle high to open SELL positions before candle low
         - candle low to open BUY positions before candle high

        Arguments:
            epic: Epic to search trades
            high: candle highest tick
            low: candle lowest tick

        """
        for trade in reversed(self.open_trades):
            if trade.epic == epic:
                if trade.direction > 0:
                    trade.on_new_tick(low)
                    trade.on_new_tick(high)
                else:
                    trade.on_new_tick(high)
                    trade.on_new_tick(low)

    ##################################################
    # CLOSE TRADE
    ##################################################
    def close_trade_api_call(self, trade: Trade, quantity: int) -> Trade:
        """
        Method used by providers trade manager to close trade by api.

        Arguments:
            trade: trade to close
            quantity: quantity to close on trade

        Returns:
            trade: updated trade after API call
        """

        return trade

    def close_trade(
        self,
        trade: Trade,
        quantity: int = None,
        close_reason: str = 'close unique trade',
        tick=None,
    ) -> None:
        """
        method to close a trade.

        This method fire the following events:

         - trade_manager_before_close_trade
         - trade_manager_after_close_trade

        Arguments:
            trade: trade to close
            quantity: quantity to close on trade (if not filled, the trade will be
            totally closed)
            close_reason: closing trade motivation (free string)
            tick: `estrade.Tick` to close the trade with (if not provided, the
                last tick received by epic will be applied)

        """
        logger.info('close unique trade %s', trade)
        quantity = (
            quantity
            if quantity and quantity <= trade.opened_quantity
            else trade.opened_quantity
        )

        self.fire(
            'trade_manager_before_close_trade',
            trade=trade,
            quantity=quantity,
            close_reason=close_reason,
        )

        updated_trade = self.market.provider.close_trade(trade, quantity)
        if not updated_trade:
            logger.error(
                'Impossible to close %d on trade because api call failed for %s'
                % (quantity, trade.ref)
            )
            return None

        updated_trade.close(quantity=quantity, close_reason=close_reason, tick=tick)

        # drop trade from open_trades and add it to closed trades
        self.open_trades.remove(updated_trade)
        self.closed_trades.append(updated_trade)

        if updated_trade.strategy:
            self.strategy_trades[updated_trade.strategy.ref]['opened'].remove(
                updated_trade
            )
            self.strategy_trades[updated_trade.strategy.ref]['closed'].append(
                updated_trade
            )

        self.fire('trade_manager_after_close_trade', trade=updated_trade)

    def close_trade_by_ref(
        self,
        ref: str,
        close_reason: str = 'close trade by ref',
        quantity: int = None,
        strategy: 'Strategy' = None,
        tick: 'Tick' = None,
    ) -> None:
        """
        Close a trade by its reference

        Arguments:
            ref: reference of the `estrade.Trade` to close
            quantity: quantity to close on trade (if not filled, the trade will be
            totally closed)
            close_reason: motivation of the trade close (free of use)
            strategy: Strategy to restrict trade search on
            tick: `estrade.Tick` to close the trade with (if not provided, the
                last tick received by epic will be applied)
        """
        for trade in reversed(self.get_trades(only_opened=True, strategy=strategy)):
            if (
                not trade.closed
                and trade.ref == ref
                and (not strategy or trade.strategy == strategy)
            ):
                self.close_trade(
                    trade, quantity=quantity, close_reason=close_reason, tick=tick
                )
                return
        raise TradeManagerException(
            f'impossible to find trade to close with ref {ref} '
            f'in {[trade.ref for trade in self.trades if not trade.closed]}'
        )

    def close_all_trades(
        self,
        close_reason: str = 'close all trades',
        strategy: 'Strategy' = None,
        tick: 'Tick' = None,
    ) -> None:
        """
        Close all open trades

        Arguments:
            close_reason: motivation of the trade close (free of use)
            strategy: Strategy to restrict trade search on
            tick: `estrade.Tick` to close the trade with (if not provided, the
                last tick received by epic will be applied)
        """
        logger.info(
            'close all (%d) trades for strategy %s'
            % (len(self.open_trades), strategy.ref if strategy else None)
        )
        for trade in self.open_trades[:]:
            if not strategy or trade.strategy == strategy:
                self.close_trade(trade, close_reason=close_reason, tick=tick)
                logger.debug('after close trade %s' % trade)
            else:
                logger.debug('do not close trade %s' % trade)

        logger.debug('%d trades remaining in open_trades' % len(self.open_trades))

    def close_all_buys(self, close_reason='close all buys', strategy=None, tick=None):
        """
        Close all opened buy trades.

        Arguments:
            close_reason: motivation of the trade close (free of use)
            strategy: Strategy to restrict trade search on
            tick: `estrade.Tick` to close the trade with (if not provided, the
                last tick received by epic will be applied)
        """
        for trade in self.open_trades[:]:
            if (
                not trade.closed
                and trade.direction > 0
                and (not strategy or trade.strategy == strategy)
            ):
                self.close_trade(trade, close_reason=close_reason, tick=tick)

    def close_all_sells(self, close_reason='close all sells', strategy=None, tick=None):
        """
        Close all opened SELL trades.

        Arguments:
            close_reason: motivation of the trade close (free of use)
            strategy: Strategy to restrict trade search on
            tick: `estrade.Tick` to close the trade with (if not provided, the
                last tick received by epic will be applied)
        """
        for trade in self.open_trades[:]:
            if (
                not trade.closed
                and trade.direction < 0
                and (not strategy or trade.strategy == strategy)
            ):
                self.close_trade(trade, close_reason=close_reason, tick=tick)

    ##################################################
    # STATS
    ##################################################
    def get_trades(
        self,
        only_opened: Optional[bool] = False,
        only_closed: Optional[bool] = False,
        strategy: Optional['Strategy'] = None,
        open_from: Optional[Union[datetime, arrow.Arrow]] = None,
        open_to: Optional[Union[datetime, arrow.Arrow]] = None,
    ) -> List[Trade]:
        """
        Get a list of trade based on search input params

        Arguments:
            only_closed: only return closed trades
            only_opened: only return opened trades
            strategy: only return trades for this strategy
            open_from: only return trades opened after this date
            open_to: only return trades before this date

        Returns:
            List of trades

        """
        trades = []
        if strategy:
            if only_closed:
                trades = self.strategy_trades[strategy.ref]['closed']
            elif only_opened:
                trades = self.strategy_trades[strategy.ref]['opened']
            else:
                trades = (
                    self.strategy_trades[strategy.ref]['opened']
                    + self.strategy_trades[strategy.ref]['closed']
                )
        else:
            if only_closed:
                trades = self.closed_trades
            elif only_opened:
                trades = self.open_trades
            else:
                trades = self.trades

        result_trades = []
        for trade in trades:
            if (open_from is None or open_from <= trade.open_tick.datetime) and (
                open_to is None or open_to >= trade.open_tick.datetime
            ):
                result_trades.append(trade)

        return result_trades

    def won_trades(self, **kwargs) -> List[Trade]:
        """
        Return list of trade with a positive (or zero) result

        Arguments:
            kwargs: see `estrade.TradeManager.get_trades` arguments.

        Returns:
            list of trade with a positive result.

        """
        return [trade for trade in self.get_trades(**kwargs) if trade.result >= 0]

    def lost_trades(self, **kwargs) -> List[Trade]:
        """
        Return list of trade with a negative result

        Arguments:
            kwargs: see `estrade.TradeManager.get_trades` arguments.

        Returns:
            list of trades with a negative result

        """
        return [trade for trade in self.get_trades(**kwargs) if trade.result < 0]

    def nb_trades(self, **kwargs) -> int:
        """
        Count number of trade matching input arguments

        Arguments:
            kwargs: see `estrade.TradeManager.get_trades` arguments.

        Returns:
            Count of trades matching input arguments.
        """
        return len(self.get_trades(**kwargs))

    def nb_win(self, **kwargs) -> int:
        """
        Count number of trade with a positive result and matching input parameters

        Arguments:
            kwargs: see `estrade.TradeManager.get_trades` arguments.

        Returns:
            Count of trades matching input arguments and having a positive result.
        """
        return len(self.won_trades(**kwargs))

    def nb_losses(self, **kwargs) -> int:
        """
        Count number of trade with a negative result and matching input parameters

        Arguments:
            kwargs: see `estrade.TradeManager.get_trades` arguments.

        Returns:
            Count of trades matching input arguments and having a positive result.
        """
        return len(self.lost_trades(**kwargs))

    def total_win(self, **kwargs) -> float:
        """
        Return the total result of trades with a positive result

        Arguments:
            kwargs: see `estrade.TradeManager.get_trades` arguments.

        Returns:
            Sum of results of trades having a positive result and matching input
                arguments.
        """
        return sum([trade.result for trade in self.won_trades(**kwargs)])

    def total_losses(self, **kwargs) -> float:
        """
        Return the total result of trades with a negative result

        Arguments:
            kwargs: see `estrade.TradeManager.get_trades` arguments.

        Returns:
            Sum of results of trades having a negative result and matching input
                arguments.
        """
        return sum([trade.result for trade in self.lost_trades(**kwargs)])

    def result(self, **kwargs) -> float:
        """
        Return the total result of trades matching input arguments

        Arguments:
            kwargs: see `estrade.TradeManager.get_trades` arguments.

        Returns:
            Sum of results of trades matching input arguments.
        """
        return sum([trade.result for trade in self.get_trades(**kwargs)])

    def profit_factor(self, **kwargs) -> float:
        """
        Return the profit factor of trades matching input arguments

        Arguments:
            kwargs: see `estrade.TradeManager.get_trades` arguments.

        Returns:
            Profit factor of the trades matching the input arguments.

        Note:
            profit factor is absolute value of (
                sum of result of trades with positive result /
                sum of results of trades with negative results
            )
        """
        total_win = self.total_win(**kwargs)
        total_losses = self.total_losses(**kwargs)

        if total_losses == 0:
            return total_win
        return total_win / abs(total_losses)

    def max_drowdown(self, **kwargs) -> float:
        """
        Max Drowdown of a list of trades

        Arguments:
            kwargs: see `estrade.TradeManager.get_trades` arguments.

        Returns:
            Max Drawdown of the trades matching the input arguments.
        """
        max_dd = 0
        min = 0
        max = 0
        result = 0
        for trade in self.get_trades(**kwargs):
            result += trade.result
            if result > max:
                max = result
                min = result
            elif result < min:
                min = max - result
                if min > max_dd:
                    max_dd = min
        return max_dd



import logging
from typing import Any, Dict, List, Optional, TYPE_CHECKING

import arrow

from estrade.candle_set import CandleSet
from estrade.candle import Candle
from estrade.mixins.ref_mixin import RefMixin
from estrade.mixins.candle_set_indicator_mixin import CandleSetIndicatorMixin
from estrade.epic import Epic
from estrade.tick import Tick
from estrade.trade import Trade
from estrade.exceptions import StrategyException

if TYPE_CHECKING:
    from estrade.market import Market

logger = logging.getLogger(__name__)


class Strategy(RefMixin):
    """
    This class define a strategy to apply on ticks/candles.t pu

    Arguments:
        epics: list of `estrade.Epic` instances
        meta: free of use dict of parameters you can re-use in your strategy methods
        ref: strategy name
        max_concurrent_trades: nb max of trades concurrently opened

    Define your custom strategies to define when to open/close trades. Your strategies
    are triggered by all new ticks received and all new candles created on the
    Epics attached to it.

    !!! example
        Example of a strategy that

         - open a BUY trade when the tick value is 1000
         - open a SELL trade when the last closed candle was red
         - close all opened BUY trades when the moving average on 50 periods is
         inferior to the current tick - 20
         - close all trades when the strategy result is > 200 when a new candle opens.

        ```python
        from estrade import CandleSet, Market, Provider, SimpleMovingAverage, Strategy

        class MyCustomStrategy(Strategy):
            def on_new_tick_opening_strategy(tick):
                if tick.value == 1000:
                    self.open_trade(
                        epic=tick.epic,
                        quantity=2,
                        direction='SELL',
                        stop_absolute=1010,
                        limit_absolute=990,
                    )

            def on_new_candle_opening_strategy(candle_set):
                last_candle = candle_set.last_closed_candle
                # check if is defined because on first candle in candle set,
                # it will be none.
                if last_candle:
                    if last_candle.color == 'red':
                        self.open_trade(
                            epic=candle_set.epic,
                            quantity=1,
                            direction='BUY',
                            stop_relative=30,
                        )

            def on_new_tick_closing_strategy(tick):
                mm50_15mn = self.get_indicator(
                    epic_ref='MY_EPIC',
                    timeframe='15minutes',
                    indicator_name='mm50'
                )
                # check if exists because for the 49 fist candles it will not be defined
                if mm50_15mn:
                    if mm50_15mn < (tick.value - 20):
                        self.close_all_buys()


            def on_new_candle_closing_strategy(tick):
                result = self.market.trade_manager.result(strategy=self)
                if result > 200:
                    self.close_all_trades()

        class MyProvider(Provider):
            # define your custom provider
            pass

        if __name__ == '__main__':
            market = Market()

            epic = Epic(ref='MY_EPIC', market=market)
            mm50 = SimpleMovingAverage(ref='mm50')
            candle_set = CandleSet(epic=epic, timeframe='15minutes', indicators=[mm50])

            strategy = MyCustomStrategy(market=market, epics=[epic])
            provider = MyProvider(market=market)

            market.run()

        ```


    """

    def __init__(
        self,
        market: 'Market',
        epics: List[Epic],
        meta: Optional[Dict[str, Any]] = None,
        ref: str = None,
        max_concurrent_trades: int = 1,
    ):
        logger.info('Init new strategy creation')
        self.market: Market = market
        self.market.strategies.append(self)
        # init class ref attribute
        RefMixin.__init__(self, ref)

        self.epics = epics
        self.meta = meta or {}
        self.max_concurrent_trades = max_concurrent_trades
        self.status = "active"

        self.paused_until: Optional[arrow.Arrow] = None
        self.stopped = False

        logger.info('New strategy created %s' % self.ref)

    ##################################################
    # EPICS
    ##################################################
    @property
    def epics(self) -> List[Epic]:
        """
        Return:
            Returns list of epics attached to strategy
        """
        return self._epics

    @epics.setter
    def epics(self, epics: List[Epic]) -> None:
        """
        Attach epics on the current strategy.

        Raises:
            StrategyExeption: if input is not a list of `estrade.Epic` instances.


        """
        self._epics = []

        if not isinstance(epics, list):
            raise StrategyException('Invalid epics on strategy creation')

        for epic in epics:
            if not isinstance(epic, Epic):
                raise StrategyException('Invalid Epic')
            logger.debug('Add epic %s to strategy %s' % (epic.ref, self.ref))
            self._epics.append(epic)
            epic.strategies.append(self)

    ##################################################
    # GETTERS
    # Helpers to fetch elements in strategy execution methods
    ##################################################
    def get_epic(self, epic_ref: str) -> Epic:
        """
        Getter to find epic in strategies epic from its ref.

        Arguments:
            epic_ref: ref of an Epic in the current strategy epics.

        Raises:
            StrategyException: if epic not found in current strategy epics

        Returns:
            Epic instance

        """
        for epic in self.epics:
            if epic.ref == epic_ref:
                return epic
        raise StrategyException(f'Epic with ref {epic_ref} not found')

    def get_candle_set(self, epic_ref: str, timeframe: str) -> CandleSet:
        """
        Getter to find a candle set from epic_ref and timeframe.

        Arguments:
            epic_ref: reference of an epic attached to this strategy
            timeframe: timeframe of the `estrade.CandleSet` to return

        Raises:
            StrategyException : if epic not found on market
            EpicException : if no candle set with this timeframe found on this epic

        Return:
            A candle set object, see `estrade.CandleSet`

        """
        return self.get_epic(epic_ref).get_candle_set(timeframe)

    def get_candle(
        self, epic_ref: str, timeframe: str, offset: int = 0
    ) -> Optional[Candle]:
        """
        Getter to find a candle by its offset in a candleset

        Arguments:
            epic_ref: reference of an epic attached to this strategy
            timeframe: timeframe of the `estrade.CandleSet` to return
            offset: index of the candle to retrieve

        !!! info "Offset usage"
            - 0: returns the currently opened candle on candleset
            - 1 : returns the last closed candle
            - etc.

        !!! tip
            Reserve usage of this getter when you need to find candles that are not the
            current candle or the last closed candle.

            It is recommendend that you call the `Strategy.get_candle_set` method and
            retrieve its `last_closed_candle`, `current_candle` for these cases.

        Raises:
            StrategyException : if epic not found on market
            EpicException : if no candle set with this timeframe found on this epic

        Return:
            A candle object, see `estrade.Candle`

        """
        offset = abs(offset) + 1
        cs = self.get_candle_set(epic_ref, timeframe)
        if len(cs.candles) >= offset:
            return cs.candles[offset * -1]
        return None

    def get_indicator(
        self, epic_ref: str, timeframe: str, indicator_name: str, offset: int = 0
    ) -> Optional[CandleSetIndicatorMixin]:
        """
        Getter to find a candle indicator value.

        Arguments:
            epic_ref: ref of an epic in current strategy epics
            timeframe: matching a candleset.timeframe in epic
            indicator_name: matching an existing indicator on candle_set
            offset: offset of the candle to fetch indicator from
                (see `Strategy.get_candle` offset argument)

        Raise:
            StrategyException : if epic not found on market
            EpicException : if no candle set with this timeframe found on this epic

        Returns:
            Indicator value
        """
        candle = self.get_candle(epic_ref, timeframe, offset)
        if not candle:
            return None
        return candle.indicators.get(indicator_name)

    def get_trades(self, **kwargs) -> Optional[List[Trade]]:
        """
        Getter to find list of strategy trades.

        !!! seealso
            referer to `TradeManager.get_trades` to view the list of arguments

        Returns:
            List of trades attached to the current strategy
        """
        return self.market.trade_manager.get_trades(strategy=self, **kwargs)

    ##################################################
    # OPEN/CLOSE trade(s) helpers
    ##################################################
    def open_trade(self, **kwargs) -> None:
        """
        Use this method to open a trade.

        This method calls `estrade.trade_manager.TradeManager.open_trade` method
        and inject this strategy as an argument.

        As a result, a new `estrade.Trade` instance is created and attached to the
        TradeManager.

        for example, you can open a trade in your strategy event as following ::

        !!! example
            Example of trade open that you can use in opening strategies
            ```python
                from estrade import Strategy

                class MyCustomStrategy:
                    def on_new_tick_opening_strategy(self, tick):
                        if tick.value == 1000:
                            self.open_trade(
                                # mandatory params
                                epic=tick.epic,
                                quantity=5,
                                direction='SELL',
                                # optional params
                                ref='my_trade_ref',
                                stop_relative=10,
                                limit_relative=20,
                                meta={'data': 'my_data'},
                            )
            ```
        """
        self.market.trade_manager.open_trade(strategy=self, **kwargs)

    def close_trade_by_ref(self, ref: str, **kwargs) -> None:
        """
        Use this method in your closing strategy methods to close a trade by its
        reference.

        !!! info
            Closing strategy method where it is advised to call this method are:

            - `Strategy.on_new_tick_closing_strategy`
            - `Strategy.on_new_candle_closing_strategy`

        !!! seealso
            This method calls `estrade.TradeManager.close_trade_by_ref` method
            and inject this strategy as an argument.

        Arguments:
            ref: existing `estrade.trade.Trade` ref code attached to this strategy on
                current Market.trade_manager
            kwargs: args to close a trade
                (see `estrade.TradeManager.close_trade_by_ref`)
        hello
        """
        self.market.trade_manager.close_trade_by_ref(ref=ref, strategy=self, **kwargs)

    def close_all_trades(self, *args, **kwargs) -> None:
        """
        Use this method in your closing strategy methods to close all opened trades
        for strategy

        !!! info
            Closing strategy method where it is advised to call this method are:

            - `Strategy.on_new_tick_closing_strategy`
            - `Strategy.on_new_candle_closing_strategy`

        !!! seealso
            This method calls `estrade.TradeManager.close_all_trades` method and inject
            this strategy as an argument.

        Arguments:
            args: args to close a trade (see `estrade.TradeManager.close_all_trades`)
            kwargs: args to close a trade (see `estrade.TradeManager.close_all_trades`)
        """
        kwargs.pop('strategy', None)
        self.market.trade_manager.close_all_trades(*args, strategy=self, **kwargs)

    def close_all_buys(self, *args, **kwargs) -> None:
        """
        Use this method in closing strategy methods to close all opened buys for
        strategy

        !!! info
            Closing strategy method where it is advised to call this method are:

            - `Strategy.on_new_tick_closing_strategy`
            - `Strategy.on_new_candle_closing_strategy`

        !!! seealso
            This method calls `estrade.TradeManager.close_all_buys` method and inject
            this strategy as an argument.

        Arguments:
            args: args to close a trade (see `estrade.TradeManager.close_all_buys`)
            kwargs: args to close a trade (see `estrade.TradeManager.close_all_buys`)
        """
        kwargs.pop('strategy', None)
        self.market.trade_manager.close_all_buys(*args, strategy=self, **kwargs)

    def close_all_sells(self, *args, **kwargs) -> None:
        """
        Use this method in closing strategy methods to close all opened sells for
        strategy

        !!! info
            Closing strategy method where it is advised to call this method are:

            - `Strategy.on_new_tick_closing_strategy`
            - `Strategy.on_new_candle_closing_strategy`

        !!! seealso
            This method calls `estrade.TradeManager.close_all_sells` method and inject
            this strategy as an argument.

        Arguments:
            args: args to close a trade (see `estrade.TradeManager.close_all_sells`)
            kwargs: args to close a trade (see `estrade.TradeManager.close_all_sells`)
        """
        kwargs.pop('strategy', None)
        self.market.trade_manager.close_all_sells(*args, strategy=self, **kwargs)

    ##################################################
    # Check if strategy method has to be applied
    ##################################################
    def _apply_closing_strategy(self):
        """
        Check if closing strategy must be applied.
        Closing strategy is applied if trades are opened.
        :return: <bool> True if closing strategy have to be applied, else False
        """
        if self.market.trade_manager.strategy_trades[self.ref]["opened"]:
            return True
        return False

    def _apply_opening_strategy(self):
        """
        Check if opening strategy must be applied.
        Opening strategy have to be applied if the number of open trades is inferior to
        the max concurrent trades
        :return: <bool> True if opening strategy have to be applied, else False
        """
        if (
            len(self.market.trade_manager.strategy_trades[self.ref]["opened"])
            < self.max_concurrent_trades
        ):
            return True
        return False

    ##################################################
    # EVENTS
    ##################################################
    def _is_strategy_paused(self, current_datetime: arrow.Arrow) -> bool:
        """
        Check if strategy is paused for a given tick datetime.
        """
        if self.paused_until is None:
            return False
        elif self.paused_until <= current_datetime:
            logger.info('Un-pause strategy %s' % self.ref)
            self.paused_until = None
            return False
        return True

    def on_new_tick(self, tick: Tick) -> None:
        """
        Method triggered when a new tick is received by Market.
        This method check if opening and/or closing srategies have to be applied and
        then call the opening/closing
        methods.

        Arguments:
            tick: new tick instance received

        """
        logger.debug('check if any trade has to be open/closed on new tick')

        self.on_every_new_tick(tick)
        apply_closing_strategy = self._apply_closing_strategy()

        if apply_closing_strategy:
            self.on_new_tick_closing_strategy(tick)

        for candle_set in tick.epic.candle_sets:
            if candle_set.new_candle_opened:
                self.on_every_new_candle(candle_set)
                if apply_closing_strategy:
                    self.on_new_candle_closing_strategy(candle_set=candle_set)

        if self._is_strategy_paused(tick.datetime):
            logger.debug('Strategy %s is paused, do not apply tick.')
            return

        if self._apply_opening_strategy():
            logger.debug('apply opening strategy')
            self.on_new_tick_opening_strategy(tick)

            for candle_set in tick.epic.candle_sets:
                if candle_set.new_candle_opened:
                    logger.debug(
                        "call new candle opening strategy bc "
                        "a new candle was created in %s" % candle_set.timeframe
                    )
                    self.on_new_candle_opening_strategy(candle_set=candle_set)

    ##################################################
    # Strategy apply method to be implemented
    ##################################################
    def on_every_new_tick(self, tick: Tick) -> None:
        """
        This method is called on every new tick received, it should be reserved
        to pause strategy (set of `self.pause_until`) or stopping strategy
        (set of `self.stopped`).
        """
        pass

    def on_every_new_candle(self, candle_set: CandleSet) -> None:
        """
        This method is called on every new tick received, it should be reserved
        to pause strategy (set of `self.pause_until`) or stopping strategy
        (set of `self.stopped`).
        """
        pass

    def set_stop(self, current_datetime: arrow.Arrow) -> None:
        """
        This method is called before every opening and closing strategy methods
        execution. Its purpose is to update the `self.stopped` parameter
        with `True` when the strategy have to be totally stopped.

        Arguments:
            current_datetime: datetime of the last received tick


        !!! note
            When the `self.stopped` value is set to `True` the strategy is totally
            deactivated. No tick will every be sent again to this strategy.

        !!! note
            If all strategy are stopped, the market run will automatically end.

        """
        pass

    def on_new_tick_opening_strategy(self, tick: Tick) -> None:
        """
        This method is called on every tick received from provider when the number of
        open trades for this strategy
        is inferior to this instance `max_concurrent_trades` value.

        Arguments:
            tick: new tick received
        """
        pass

    def on_new_tick_closing_strategy(self, tick: Tick) -> None:
        """
        Method is called on every tick received from provider when at least one trade is
        opened for this strategy.

        Arguments:
            tick: new tick received

        """
        pass

    def on_new_candle_opening_strategy(self, candle_set: CandleSet) -> None:
        """
        Method called every time a new candle is created when the number of opened
        trades for this strategy is inferior to this instance `max_concurrent_trades`
        value.

        Use this method to open trades with `Strategy.open_trade` method.

        Arguments:
            candle_set: candle set that generated a new candle

        """
        pass

    def on_new_candle_closing_strategy(self, candle_set: CandleSet) -> None:
        """
        Method called every time a new candle is created and at least one trade is
        opened for this strategy.

        Arguments:
            candle_set: candle set that generated a new candle

        """
        pass

    ##################################################
    # Reporting
    ##################################################
    @property
    def to_json(self) -> Dict[str, Any]:
        """
        Generate a dictionary of strategy stats for reporting
        """
        strategy_stats = {
            "ref": self.ref,
            "max_concurrent_trades": self.max_concurrent_trades,
            "profit_factor": None,
            "result": None,
            "nb_trades": 0,
            "nb_win": 0,
            "nb_losses": 0,
            **self.meta,
        }
        if self.market and self.market.trade_manager:
            strategy_stats["profit_factor"] = self.market.trade_manager.profit_factor(
                strategy=self
            )
            strategy_stats["result"] = self.market.trade_manager.result(strategy=self)
            strategy_stats["nb_trades"] = self.market.trade_manager.nb_trades(
                strategy=self
            )
            strategy_stats["nb_win"] = self.market.trade_manager.nb_win(strategy=self)
            strategy_stats["nb_losses"] = self.market.trade_manager.nb_losses(
                strategy=self
            )
            strategy_stats["max_drawdown"] = self.market.trade_manager.max_drowdown(
                strategy=self
            )

        return strategy_stats

    @property
    def json_headers(self) -> List[str]:
        """
        Generate header of a dictionary of strategy stats for reporting
        """
        headers = [
            "ref",
            "profit_factor",
            "result",
            "nb_trades",
            "nb_win",
            "nb_losses",
            "max_drawdown",
            "max_concurrent_trades",
        ]
        headers.extend(list(self.meta.keys()))
        return headers
